@@ -78,8 +78,6 @@ aggregator_new(
 	ret->computes = NULL;
 	ret->next = NULL;
 
-	pthread_mutex_init(&ret->bucketlock, NULL);
-
 	return ret;
 }
 
@@ -143,6 +141,7 @@ aggregator_add_compute(
 	ac->metric = strdup(metric);
 	memset(ac->invocations_ht, 0, sizeof(ac->invocations_ht));
 	ac->entries_needed = store;
+	pthread_rwlock_init(&ac->invlock, NULL);
 	ac->next = NULL;
 
 	return 0;
@@ -197,7 +196,7 @@ aggregator_putmetric(
 	/* get value */
 	if ((v = strchr(firstspace + 1, ' ')) == NULL) {
 		/* metric includes \n */
-		if (mode == DEBUG || mode == DEBUGTEST)
+		if (mode & MODE_DEBUG)
 			logerr("aggregator: dropping incorrect metric: %s",
 					metric);
 		return;
@@ -208,7 +207,6 @@ aggregator_putmetric(
 	val = atof(firstspace + 1);
 	epoch = atoll(v + 1);
 
-	pthread_mutex_lock(&s->bucketlock);
 	for (compute = s->computes; compute != NULL; compute = compute->next) {
 		if (nmatch == 0) {
 			ometric = compute->metric;
@@ -232,11 +230,26 @@ aggregator_putmetric(
 		omhtbucket =
 			((omhash >> AGGR_HT_POW_SIZE) ^ omhash) &
 			(((unsigned int)1 << AGGR_HT_POW_SIZE) - 1);
-		invocation = compute->invocations_ht[omhtbucket];
-		for (; invocation != NULL; invocation = invocation->next)
-			if (invocation->hash == omhash &&
-					strcmp(ometric, invocation->metric) == 0)  /* match */
+
+#define find_invocation(o) \
+		invocation = compute->invocations_ht[omhtbucket]; \
+		for (; invocation != NULL; invocation = invocation->next) \
+			if (invocation->hash == omhash && \
+					strcmp(o, invocation->metric) == 0)  /* match */ \
 				break;
+		pthread_rwlock_rdlock(&compute->invlock);
+		find_invocation(ometric);
+		/* switch to a write lock from here, since we've found what we
+		 * were looking for (or are going to create it) and modify it */
+		pthread_rwlock_unlock(&compute->invlock);
+		pthread_rwlock_wrlock(&compute->invlock);
+
+		if (invocation == NULL) {
+			/* we need to recheck there wasn't someone else who did the
+			 * same thing we want to do below */
+			find_invocation(ometric);
+		}
+
 		if (invocation == NULL) {  /* no match, add */
 			int i;
 			time_t now;
@@ -244,12 +257,14 @@ aggregator_putmetric(
 			if ((invocation = malloc(sizeof(*invocation))) == NULL) {
 				logerr("aggregator: out of memory creating %s from %s",
 						ometric, metric);
+				pthread_rwlock_unlock(&compute->invlock);
 				continue;
 			}
 			if ((invocation->metric = strdup(ometric)) == NULL) {
 				logerr("aggregator: out of memory creating %s from %s",
 						ometric, metric);
 				free(invocation);
+				pthread_rwlock_unlock(&compute->invlock);
 				continue;
 			}
 			invocation->hash = omhash;
@@ -276,6 +291,7 @@ aggregator_putmetric(
 						ometric, metric);
 				free(invocation->metric);
 				free(invocation);
+				pthread_rwlock_unlock(&compute->invlock);
 				continue;
 			}
 			for (i = 0; i < s->bucketcnt; i++) {
@@ -295,17 +311,19 @@ aggregator_putmetric(
 		if (itime < 0) {
 			/* drop too old metric */
 			s->dropped++;
+			pthread_rwlock_unlock(&compute->invlock);
 			continue;
 		}
 
 		slot = itime / s->interval;
 		if (slot >= s->bucketcnt) {
-			if (mode == DEBUG || mode == DEBUGTEST)
+			if (mode & MODE_DEBUG)
 				logerr("aggregator: dropping metric too far in the "
 						"future (%lld > %lld): %s from %s", epoch,
 						invocation->buckets[s->bucketcnt - 1].start,
 						ometric, metric);
 			s->dropped++;
+			pthread_rwlock_unlock(&compute->invlock);
 			continue;
 		}
 
@@ -339,8 +357,9 @@ aggregator_putmetric(
 				entries->values[bucket->cnt] = val;
 		}
 		bucket->cnt++;
+
+		pthread_rwlock_unlock(&compute->invlock);
 	}
-	pthread_mutex_unlock(&s->bucketlock);
 
 	return;
 }
@@ -370,6 +389,7 @@ aggregator_expire(void *sub)
 	double *values;
 	size_t len = 0;
 	int i;
+	size_t k;
 	unsigned char j;
 	int work;
 	char metric[METRIC_BUFSIZ];
@@ -439,10 +459,9 @@ aggregator_expire(void *sub)
 										break;
 									case MEDN:
 										/* median == 50th percentile */
-									case PCTL: {
+									case PCTL:
 										/* nearest rank method */
-										size_t n =
-											(int)(((double)c->percentile/100.0 *
+										k = (int)(((double)c->percentile/100.0 *
 														(double)b->cnt) + 0.9);
 										values = b->entries.values;
 										/* TODO: lazy approach, in case
@@ -456,16 +475,16 @@ aggregator_expire(void *sub)
 										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric,
-												values[n - 1],
+												values[k - 1],
 												ts);
-									}	break;
+										break;
 									case VAR:
 									case SDEV: {
 										double avg = b->sum / (double)b->cnt;
 										double ksum = 0;
 										values = b->entries.values;
-										for (i = 0; i < b->cnt; i++)
-											ksum += pow(values[i] - avg, 2);
+										for (k = 0; k < b->cnt; k++)
+											ksum += pow(values[k] - avg, 2);
 										ksum /= (double)b->cnt;
 										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
@@ -477,7 +496,15 @@ aggregator_expire(void *sub)
 									default:
 									   assert(0);  /* for compiler (len) */
 								}
-								if (write(s->fd, metric, len) != len) {
+								ts = write(s->fd, metric, len);
+								if (ts < 0) {
+									logerr("aggregator: failed to write to "
+											"pipe (fd=%d): %s\n",
+											s->fd, strerror(errno));
+									s->dropped++;
+								} else if (ts < len) {
+									logerr("aggregator: uncomplete write on "
+											"pipe (fd=%d)\n", s->fd);
 									s->dropped++;
 								} else {
 									s->sent++;
@@ -486,7 +513,7 @@ aggregator_expire(void *sub)
 
 							/* move the bucket to the end, to make room for
 							 * new ones */
-							pthread_mutex_lock(&s->bucketlock);
+							pthread_rwlock_wrlock(&c->invlock);
 							b = &inv->buckets[0];
 							len = b->entries.size;
 							values = b->entries.values;
@@ -499,23 +526,22 @@ aggregator_expire(void *sub)
 								s->interval;
 							b->entries.size = len;
 							b->entries.values = values;
-							pthread_mutex_unlock(&s->bucketlock);
+							pthread_rwlock_unlock(&c->invlock);
 
 							work++;
 						}
 
 						if (isempty) {
 							/* see if the remaining buckets are empty too */
-							pthread_mutex_lock(&s->bucketlock);
 							for (j = 0; j < s->bucketcnt; j++) {
 								if (inv->buckets[j].cnt != 0) {
 									isempty = 0;
-									pthread_mutex_unlock(&s->bucketlock);
 									break;
 								}
 							}
 						}
 						if (isempty) {
+							pthread_rwlock_wrlock(&c->invlock);
 							/* free and unlink */
 							if (c->entries_needed)
 								for (j = 0; j < s->bucketcnt; j++)
@@ -532,7 +558,7 @@ aggregator_expire(void *sub)
 								free(inv);
 								inv = c->invocations_ht[i];
 							}
-							pthread_mutex_unlock(&s->bucketlock);
+							pthread_rwlock_unlock(&c->invlock);
 						} else {
 							lastinv = inv;
 							inv = inv->next;
@@ -574,6 +600,8 @@ aggregator_expire(void *sub)
 				}
 			}
 
+			pthread_rwlock_destroy(&c->invlock);
+
 			s->computes = c->next;
 			free(c);
 		}
@@ -661,9 +689,8 @@ aggregator_get_received(aggregator *a)
 inline size_t
 aggregator_get_received_sub(aggregator *aggrs)
 {
-	size_t d = aggregator_get_received(aggrs);
-	size_t r = d - prevreceived;
-	prevreceived += d;
+	size_t r = aggregator_get_received(aggrs) - prevreceived;
+	prevreceived += r;
 	return r;
 }
 
@@ -688,9 +715,8 @@ aggregator_get_sent(aggregator *a)
 inline size_t
 aggregator_get_sent_sub(aggregator *aggrs)
 {
-	size_t d = aggregator_get_sent(aggrs);
-	size_t r = d - prevsent;
-	prevsent += d;
+	size_t r = aggregator_get_sent(aggrs) - prevsent;
+	prevsent += r;
 	return r;
 }
 
@@ -717,8 +743,7 @@ aggregator_get_dropped(aggregator *a)
 inline size_t
 aggregator_get_dropped_sub(aggregator *aggrs)
 {
-	size_t d = aggregator_get_dropped(aggrs);
-	size_t r = d - prevdropped;
-	prevdropped += d;
+	size_t r = aggregator_get_dropped(aggrs) - prevdropped;
+	prevdropped += r;
 	return r;
 }
