@@ -53,6 +53,7 @@ typedef struct _connection {
 	destination dests[CONN_DESTS_SIZE];
 	size_t destlen;
 	struct timeval lastwork;
+	unsigned int maxsenddelay;
 	char hadwork:1;
 	char isaggr:1;
 } connection;
@@ -64,13 +65,14 @@ struct _dispatcher {
 	size_t metrics;
 	size_t blackholes;
 	size_t ticks;
+	size_t sleeps;
 	size_t prevmetrics;
 	size_t prevblackholes;
 	size_t prevticks;
-	enum { RUNNING, SLEEPING } state;
+	size_t prevsleeps;
 	char keep_running:1;
-	route *routes;
-	route *pending_routes;
+	router *rtr;
+	router *pending_rtr;
 	char route_refresh_pending:1;
 	char hold:1;
 	char *allowed_chars;
@@ -293,9 +295,13 @@ inline static char
 dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
 {
 	int i;
-	char force = timediff(conn->lastwork, now) > 1 * 1000 * 1000;  /* 1 sec timeout */
+	char force;
 
 	if (conn->destlen > 0) {
+		if (conn->maxsenddelay == 0)
+			conn->maxsenddelay = ((rand() % 750) + 250) * 1000;
+		/* force after timeout */
+		force = timediff(conn->lastwork, now) > conn->maxsenddelay;
 		for (i = 0; i < conn->destlen; i++) {
 			if (server_send(conn->dests[i].dest, conn->dests[i].metric, force) == 0)
 				break;
@@ -338,22 +344,16 @@ dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
  *      block reads
  */
 static int
-dispatch_connection(connection *conn, dispatcher *self)
+dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 {
 	char *p, *q, *firstspace, *lastnl;
 	int len;
-	struct timeval start, stop;
 
-	gettimeofday(&start, NULL);
 	/* first try to resume any work being blocked */
 	if (dispatch_process_dests(conn, self, start) == 0) {
-		gettimeofday(&stop, NULL);
-		self->ticks += timediff(start, stop);
 		conn->takenby = 0;
 		return 0;
 	}
-	gettimeofday(&stop, NULL);
-	self->ticks += timediff(start, stop);
 
 	/* don't poll (read) when the last time we ran nothing happened,
 	 * this is to avoid excessive CPU usage, issue #126 */
@@ -363,7 +363,6 @@ dispatch_connection(connection *conn, dispatcher *self)
 	}
 	conn->hadwork = 0;
 
-	gettimeofday(&start, NULL);
 	len = -2;
 	/* try to read more data, if that succeeds, or we still have data
 	 * left in the buffer, try to process the buffer */
@@ -402,10 +401,10 @@ dispatch_connection(connection *conn, dispatcher *self)
 				*q = '\0';  /* can do this because we substract one from buf */
 
 				/* perform routing of this metric */
-				self->blackholes += router_route(
+				self->blackholes += router_route(self->rtr,
 						conn->dests, &conn->destlen, CONN_DESTS_SIZE,
 						conn->srcaddr,
-						conn->metric, firstspace, self->routes);
+						conn->metric, firstspace);
 
 				/* restart building new one from the start */
 				q = conn->metric;
@@ -413,6 +412,7 @@ dispatch_connection(connection *conn, dispatcher *self)
 
 				conn->hadwork = 1;
 				gettimeofday(&conn->lastwork, NULL);
+				conn->maxsenddelay = 0;
 				/* send the metric to where it is supposed to go */
 				if (dispatch_process_dests(conn, self, conn->lastwork) == 0)
 					break;
@@ -458,13 +458,13 @@ dispatch_connection(connection *conn, dispatcher *self)
 			memmove(conn->buf, lastnl + 1, conn->buflen);
 		}
 	}
-	gettimeofday(&stop, NULL);
-	self->ticks += timediff(start, stop);
 	if (len == -1 && (errno == EINTR ||
 				errno == EAGAIN ||
 				errno == EWOULDBLOCK))
 	{
 		/* nothing available/no work done */
+		struct timeval stop;
+		gettimeofday(&stop, NULL);
 		if (!conn->noexpire &&
 				timediff(conn->lastwork, stop) > IDLE_DISCONNECT_TIME)
 		{
@@ -514,16 +514,16 @@ dispatch_runner(void *arg)
 {
 	dispatcher *self = (dispatcher *)arg;
 	connection *conn;
-	int work;
 	int c;
 
 	self->metrics = 0;
 	self->blackholes = 0;
 	self->ticks = 0;
+	self->sleeps = 0;
 	self->prevmetrics = 0;
 	self->prevblackholes = 0;
 	self->prevticks = 0;
-	self->state = SLEEPING;
+	self->prevsleeps = 0;
 
 	if (self->type == LISTENER) {
 		struct pollfd ufds[sizeof(listeners) / sizeof(connection *)];
@@ -558,15 +558,20 @@ dispatch_runner(void *arg)
 			}
 		}
 	} else if (self->type == CONNECTION) {
+		int work;
+		struct timeval start, stop;
+
 		while (self->keep_running) {
 			work = 0;
+
 			if (self->route_refresh_pending) {
-				self->routes = self->pending_routes;
-				self->pending_routes = NULL;
+				self->rtr = self->pending_rtr;
+				self->pending_rtr = NULL;
 				self->route_refresh_pending = 0;
 				self->hold = 0;
 			}
 
+			gettimeofday(&start, NULL);
 			pthread_rwlock_rdlock(&connectionslock);
 			for (c = 0; c < connectionslen; c++) {
 				conn = &(connections[c]);
@@ -577,15 +582,19 @@ dispatch_runner(void *arg)
 					conn->takenby = 0;
 					continue;
 				}
-				self->state = RUNNING;
-				work += dispatch_connection(conn, self);
+				work += dispatch_connection(conn, self, start);
 			}
 			pthread_rwlock_unlock(&connectionslock);
+			gettimeofday(&stop, NULL);
+			self->ticks += timediff(start, stop);
 
-			self->state = SLEEPING;
 			/* nothing done, avoid spinlocking */
-			if (self->keep_running && work == 0)
+			if (self->keep_running && work == 0) {
+				gettimeofday(&start, NULL);
 				usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
+				gettimeofday(&stop, NULL);
+				self->sleeps += timediff(start, stop);
+			}
 		}
 	} else {
 		logerr("huh? unknown self type!\n");
@@ -599,7 +608,7 @@ dispatch_runner(void *arg)
  * Returns its handle.
  */
 static dispatcher *
-dispatch_new(char id, enum conntype type, route *routes, char *allowed_chars)
+dispatch_new(char id, enum conntype type, router *r, char *allowed_chars)
 {
 	dispatcher *ret = malloc(sizeof(dispatcher));
 
@@ -609,7 +618,7 @@ dispatch_new(char id, enum conntype type, route *routes, char *allowed_chars)
 	ret->id = id;
 	ret->type = type;
 	ret->keep_running = 1;
-	ret->routes = routes;
+	ret->rtr = r;
 	ret->route_refresh_pending = 0;
 	ret->hold = 0;
 	ret->allowed_chars = allowed_chars;
@@ -639,10 +648,10 @@ dispatch_new_listener(void)
  * existing connections.
  */
 dispatcher *
-dispatch_new_connection(route *routes, char *allowed_chars)
+dispatch_new_connection(router *r, char *allowed_chars)
 {
 	char id = globalid++;
-	return dispatch_new(id, CONNECTION, routes, allowed_chars);
+	return dispatch_new(id, CONNECTION, r, allowed_chars);
 }
 
 /**
@@ -655,14 +664,22 @@ dispatch_stop(dispatcher *d)
 }
 
 /**
- * Shuts down and frees up dispatcher d.  Returns when the dispatcher
- * has terminated.
+ * Shuts down dispatcher d.  Returns when the dispatcher has terminated.
  */
 void
 dispatch_shutdown(dispatcher *d)
 {
 	dispatch_stop(d);
 	pthread_join(d->tid, NULL);
+}
+
+/**
+ * Free up resources taken by dispatcher d.  The caller should make sure
+ * the dispatcher has been shut down at this point.
+ */
+void
+dispatch_free(dispatcher *d)
+{
 	free(d);
 }
 
@@ -684,9 +701,9 @@ dispatch_hold(dispatcher *d)
  * replacement is performed at the next cycle of the dispatcher.
  */
 inline void
-dispatch_schedulereload(dispatcher *d, route *r)
+dispatch_schedulereload(dispatcher *d, router *r)
 {
-	d->pending_routes = r;
+	d->pending_rtr = r;
 	d->route_refresh_pending = 1;
 }
 
@@ -721,6 +738,28 @@ dispatch_get_ticks_sub(dispatcher *self)
 	return d;
 }
 
+/**
+ * Returns the wall-clock time in milliseconds consumed while sleeping
+ * by this dispatcher.
+ */
+inline size_t
+dispatch_get_sleeps(dispatcher *self)
+{
+	return self->sleeps;
+}
+
+/**
+ * Returns the wall-clock time consumed while sleeping since last call
+ * to this function.
+ */
+inline size_t
+dispatch_get_sleeps_sub(dispatcher *self)
+{
+	size_t d = self->sleeps - self->prevsleeps;
+	self->prevsleeps += d;
+	return d;
+}
+
 /**
  * Returns the number of metrics dispatched since start.
  */
@@ -764,18 +803,6 @@ dispatch_get_blackholes_sub(dispatcher *self)
 	return d;
 }
 
-/**
- * Returns whether this dispatcher is currently running, or not.  A
- * dispatcher is running when it is actively handling a connection, and
- * all tasks related to getting the data received in the place where it
- * should be.
- */
-inline char
-dispatch_busy(dispatcher *self)
-{
-	return self->state == RUNNING;
-}
-
 /**
  * Returns the number of accepted connections thusfar.
  */
