@@ -24,8 +24,10 @@
 #include <time.h>
 #include <errno.h>
 #include <assert.h>
-#if defined(ENABLE_OPENMP)
-# include <omp.h>
+
+#if defined(__MACH__) || defined(BSD)
+# include <sys/types.h>
+# include <sys/sysctl.h>
 #endif
 
 #include "relay.h"
@@ -38,7 +40,7 @@
 
 int keep_running = 1;
 char relay_hostname[256];
-enum rmode mode = NORMAL;
+unsigned char mode = 0;
 
 static char *config = NULL;
 static int batchsize = 2500;
@@ -46,9 +48,7 @@ static int queuesize = 25000;
 static unsigned short iotimeout = 600;
 static dispatcher **workers = NULL;
 static char workercnt = 0;
-static cluster *clusters = NULL;
-static route *routes = NULL;
-static aggregator *aggrs = NULL;
+static router *rtr = NULL;
 static server *internal_submission = NULL;
 static char *relay_logfile = NULL;
 static FILE *relay_stdout = NULL;
@@ -113,24 +113,24 @@ relaylog(enum logdst dest, const char *fmt, ...)
 static void
 exit_handler(int sig)
 {
-	char *signal = "unknown signal";
+	char *sign = "unknown signal";
 
 	switch (sig) {
 		case SIGTERM:
-			signal = "SIGTERM";
+			sign = "SIGTERM";
 			break;
 		case SIGINT:
-			signal = "SIGINT";
+			sign = "SIGINT";
 			break;
 		case SIGQUIT:
-			signal = "SIGQUIT";
+			sign = "SIGQUIT";
 			break;
 	}
 	if (keep_running) {
-		logout("caught %s\n", signal);
+		logout("caught %s\n", sign);
 	} else {
 		logerr("caught %s while already shutting down, "
-				"forcing exit!\n", signal);
+				"forcing exit!\n", sign);
 		exit(1);
 	}
 	keep_running = 0;
@@ -139,14 +139,11 @@ exit_handler(int sig)
 static void
 hup_handler(int sig)
 {
-	route *newroutes;
-	cluster *newclusters;
+	router *newrtr;
 	aggregator *newaggrs;
 	int id;
 	FILE *newfd;
 	size_t numaggregators;
-	server **servers;
-	size_t i;
 
 	logout("caught SIGHUP...\n");
 	if (relay_stderr != stderr) {
@@ -167,16 +164,16 @@ hup_handler(int sig)
 	}
 
 	logout("reloading config from '%s'\n", config);
-	if (router_readconfig(&newclusters, &newroutes, &newaggrs,
-				config, queuesize, batchsize, iotimeout) == 0)
+	if ((newrtr = router_readconfig(config,
+					queuesize, batchsize, iotimeout)) == NULL)
 	{
 		logerr("failed to read configuration '%s', aborting reload\n", config);
 		return;
 	}
-	router_optimise(&newroutes);
+	router_optimise(newrtr);
 
 	logout("reloading collector\n");
-	collector_schedulereload(newclusters, newaggrs);
+	collector_schedulereload(newrtr);
 	while (!collector_reloadcomplete())
 		usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
 
@@ -192,7 +189,7 @@ hup_handler(int sig)
 	for (id = 1; id < 1 + workercnt; id++)
 		dispatch_hold(workers[id]);
 
-	numaggregators = aggregator_numaggregators(aggrs);
+	numaggregators = aggregator_numaggregators(router_getaggregators(rtr));
 	if (numaggregators > 0) {
 		logout("expiring aggregations\n");
 		aggregator_stop();  /* frees aggrs */
@@ -201,6 +198,7 @@ hup_handler(int sig)
 		 * dispatchers, which will hand over to the servers with proper
 		 * metric adjustments (stubs). */
 	}
+	newaggrs = router_getaggregators(newrtr);
 	numaggregators = aggregator_numaggregators(newaggrs);
 	if (numaggregators > 0) {
 		if (!aggregator_start(newaggrs)) {
@@ -211,21 +209,15 @@ hup_handler(int sig)
 
 	logout("reloading workers\n");
 	for (id = 1; id < 1 + workercnt; id++)
-		dispatch_schedulereload(workers[id], newroutes); /* un-holds */
+		dispatch_schedulereload(workers[id], newrtr); /* un-holds */
 	for (id = 1; id < 1 + workercnt; id++) {
 		while (!dispatch_reloadcomplete(workers[id + 0]))
 			usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
 	}
 
-	servers = router_getservers(clusters);
-	for (i = 0; servers[i] != NULL; i++)
-		server_stop(servers[i]);
-	free(servers);
-	router_free(clusters, routes);
+	router_free(rtr);
 
-	routes = newroutes;
-	clusters = newclusters;
-	aggrs = newaggrs;
+	rtr = newrtr;
 
 	logout("SIGHUP handler complete\n");
 }
@@ -233,11 +225,16 @@ hup_handler(int sig)
 static int
 get_cores(void)
 {
-#if defined(ENABLE_OPENMP)
-	return omp_get_num_procs();
-#else
-	return 5;
+	int cpus = 5;
+
+#if defined(__MACH__) || defined(BSD)
+	size_t len = sizeof(cpus);
+	sysctlbyname("hw.ncpu", &cpus, &len, NULL, 0);
+#else /* Linux and Solaris */
+	cpus = sysconf(_SC_NPROCESSORS_ONLN);
 #endif
+
+	return cpus;
 }
 
 static void
@@ -273,6 +270,8 @@ do_usage(int exitcode)
 	       "      statistics, report drop counts and queue pressure to log\n");
 	printf("  -t  config test mode: prints rule matches from input on stdin\n");
 	printf("  -H  hostname: override hostname (used in statistics)\n");
+	printf("  -D  daemonise: run in a background\n");
+	printf("  -P  pidfile: write a pid to a specified pidfile\n");
 
 	exit(exitcode);
 }
@@ -290,44 +289,31 @@ main(int argc, char * const argv[])
 	int ch;
 	size_t numaggregators;
 	char *listeninterface = NULL;
-	server **servers;
 	char *allowed_chars = NULL;
-	int i;
 	enum { SUB, CUM } smode = CUM;
+	char *pidfile = NULL;
+	FILE *pidfile_handle = NULL;
+	aggregator *aggrs = NULL;
 
 	if (gethostname(relay_hostname, sizeof(relay_hostname)) < 0)
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
-	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:S:T:c:H:B:")) != -1) {
+	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:S:T:c:H:B:DP:")) != -1) {
 		switch (ch) {
 			case 'v':
 				do_version();
 				break;
 			case 'd':
-				if (mode == TEST) {
-					mode = DEBUGTEST;
-				} else if (mode == SUBMISSION) {
-					mode = DEBUGSUBMISSION;
-				} else {
-					mode = DEBUG;
-				}
+				mode |= MODE_DEBUG;
 				break;
 			case 'm':
 				smode = SUB;
 				break;
 			case 's':
-				if (mode == DEBUG) {
-					mode = DEBUGSUBMISSION;
-				} else {
-					mode = SUBMISSION;
-				}
+				mode |= MODE_SUBMISSION;
 				break;
 			case 't':
-				if (mode == DEBUG) {
-					mode = DEBUGTEST;
-				} else {
-					mode = TEST;
-				}
+				mode |= MODE_TEST;
 				break;
 			case 'f':
 				config = optarg;
@@ -399,6 +385,13 @@ main(int argc, char * const argv[])
 				}
 				listenbacklog = (unsigned int)val;
 			}	break;
+
+			case 'D':
+				mode |= MODE_DAEMON;
+				break;
+			case 'P':
+				pidfile = optarg;
+				break;
 			case '?':
 			case ':':
 				do_usage(1);
@@ -417,7 +410,7 @@ main(int argc, char * const argv[])
 	srand(time(NULL));
 
 	if (workercnt == 0)
-		workercnt = (mode == SUBMISSION || mode == DEBUGSUBMISSION) ? 2 : get_cores();
+		workercnt = mode & MODE_SUBMISSION ? 2 : get_cores();
 
 	/* any_of failover maths need batchsize to be smaller than queuesize */
 	if (batchsize > queuesize) {
@@ -425,7 +418,7 @@ main(int argc, char * const argv[])
 		exit(-1);
 	}
 
-	if (relay_logfile != NULL && mode != TEST && mode != DEBUGTEST) {
+	if (relay_logfile != NULL && (mode & MODE_TEST) == 0) {
 		FILE *f = fopen(relay_logfile, "a");
 		if (f == NULL) {
 			fprintf(stderr, "error: failed to open logfile '%s': %s\n",
@@ -440,6 +433,104 @@ main(int argc, char * const argv[])
 	}
 	relay_can_log = 1;
 
+	if (mode & MODE_DAEMON) {
+		if (relay_logfile == NULL) {
+			fprintf(stderr,
+					"You must specify logfile if you want daemonisation!\n");
+			exit(-1);
+		}
+
+		if (mode & MODE_TEST) {
+			fprintf(stderr,
+					"You cannot use test mode if you want daemonisation!\n");
+			exit(-1);
+		}
+	}
+
+	if (pidfile != NULL) {
+		pidfile_handle = fopen(pidfile, "w");
+		if (pidfile_handle == NULL) {
+			fprintf(stderr, "failed to open pidfile '%s': %s\n",
+					pidfile, strerror(errno));
+			exit(-1);
+		}
+	}
+
+	if (mode & MODE_DAEMON) {
+		pid_t p;
+		int fds[2];
+
+		if (pipe(fds) < 0) {
+			fprintf(stderr, 
+					"failed to create pipe to talk to child process: %s\n",
+					strerror(errno));
+			exit(1);
+		}
+
+		p = fork();
+		if (p < 0) {
+			fprintf(stderr, "failed to fork intermediate process: %s\n",
+					strerror(errno));
+			exit(1);
+		} else if (p == 0) {
+			char msg[1024];
+			/* child, for again so our intermediate process can die and
+			 * its child becomes an orphan to init */
+			p = fork();
+			if (p < 0) {
+				snprintf(msg, sizeof(msg), "failed to fork daemon process: %s",
+						strerror(errno));
+				write(fds[1], msg, strlen(msg) + 1);
+				exit(1); /* not that it matters */
+			} else if (p == 0) {
+				/* child, this is the final daemon process */
+				if (setsid() < 0) {
+					snprintf(msg, sizeof(msg), "failed to setsid(): %s",
+							strerror(errno));
+					write(fds[1], msg, strlen(msg) + 1);
+					exit(1); /* not that it matters */
+				}
+				/* close descriptors, we won't need them ever since
+				 * we're detached from the controlling terminal */
+				close(0);
+				close(1);
+				close(2);
+
+				/* we're fine, flag our grandparent (= the original
+				 * process being called) it can terminate */
+				write(fds[1], "OK", 3);
+			} else {
+				/* parent, die */
+				exit(0);
+			}
+		} else {
+			/* parent: wait for child to report success */
+			char msg[1024];
+			int len = read(fds[0], msg, sizeof(msg));
+			if (len < 0) {
+				fprintf(stderr, "unable to read from child, "
+						"assuming daemonisation failed\n");
+				exit(1);
+			} else if (len == 0) {
+				fprintf(stderr, "got unexpected EOF from child, "
+						"assuming daemonisation failed\n");
+				exit(1);
+			} else if (strncmp(msg, "OK", len) == 0) {
+				/* child is happy, so can we */
+				exit(0);
+			} else {
+				fprintf(stderr, "%s\n", msg);
+				exit(1);
+			}
+		}
+	}
+
+	if (pidfile_handle) {
+		fprintf(pidfile_handle, "%d\n", getpid());
+		fclose(pidfile_handle);
+	}
+
+
 	logout("starting carbon-c-relay v%s (%s), pid=%d\n",
 			VERSION, GIT_VERSION, getpid());
 	fprintf(relay_stdout, "configuration:\n");
@@ -458,37 +549,40 @@ main(int argc, char * const argv[])
 	if (allowed_chars != NULL)
 		fprintf(relay_stdout, "    extra allowed characters = %s\n",
 				allowed_chars);
-	if (mode == DEBUG || mode == DEBUGTEST || mode == DEBUGSUBMISSION)
+	if (mode & MODE_DEBUG)
 		fprintf(relay_stdout, "    debug = true\n");
-	else if (mode == SUBMISSION || mode == DEBUGSUBMISSION)
+	else if (mode & MODE_SUBMISSION)
 		fprintf(relay_stdout, "    submission = true\n");
+	else if (mode & MODE_DAEMON)
+		fprintf(relay_stdout, "    daemon = true\n");
 	fprintf(relay_stdout, "    routes configuration = %s\n", config);
 	fprintf(relay_stdout, "\n");
 
-	if (router_readconfig(&clusters, &routes, &aggrs,
-				config, queuesize, batchsize, iotimeout) == 0)
+	if ((rtr = router_readconfig(config,
+					queuesize, batchsize, iotimeout)) == NULL)
 	{
 		logerr("failed to read configuration '%s'\n", config);
 		return 1;
 	}
-	router_optimise(&routes);
+	router_optimise(rtr);
 
+	aggrs = router_getaggregators(rtr);
 	numaggregators = aggregator_numaggregators(aggrs);
-#define dbg (mode == DEBUG || mode == DEBUGTEST ? 2 : 0)
-	if (numaggregators > 10 && !dbg) {
+	if (numaggregators > 10 && mode & MODE_DEBUG) {
 		fprintf(relay_stdout, "parsed configuration follows:\n"
 				"(%zu aggregations with %zu computations omitted "
 				"for brevity)\n",
 				numaggregators, aggregator_numcomputes(aggrs));
-		router_printconfig(relay_stdout, 0, clusters, routes);
+		router_printconfig(rtr, relay_stdout, 0);
 	} else {
 		fprintf(relay_stdout, "parsed configuration follows:\n");
-		router_printconfig(relay_stdout, 1 + dbg, clusters, routes);
+		router_printconfig(rtr, relay_stdout,
+				1 + (mode & MODE_DEBUG ? 2 : 0));
 	}
 	fprintf(relay_stdout, "\n");
 
 	/* shortcut for rule testing mode */
-	if (mode == TEST || mode == DEBUGTEST) {
+	if (mode & MODE_TEST) {
 		char metricbuf[METRIC_BUFSIZ];
 		char *p;
 
@@ -496,7 +590,7 @@ main(int argc, char * const argv[])
 		while (fgets(metricbuf, sizeof(metricbuf), stdin) != NULL) {
 			if ((p = strchr(metricbuf, '\n')) != NULL)
 				*p = '\0';
-			router_test(metricbuf, routes);
+			router_test(rtr, metricbuf);
 		}
 
 		exit(0);
@@ -556,7 +650,7 @@ main(int argc, char * const argv[])
 		allowed_chars = "-_:#";
 	logout("starting %d workers\n", workercnt);
 	for (id = 1; id < 1 + workercnt; id++) {
-		workers[id + 0] = dispatch_new_connection(routes, allowed_chars);
+		workers[id + 0] = dispatch_new_connection(rtr, allowed_chars);
 		if (workers[id + 0] == NULL) {
 			logerr("failed to add worker %d\n", id);
 			break;
@@ -586,8 +680,7 @@ main(int argc, char * const argv[])
 	}
 
 	logout("starting statistics collector\n");
-	collector_start(&workers[1], clusters, aggrs, internal_submission,
-			smode == CUM);
+	collector_start(&workers[1], rtr, internal_submission, smode == CUM);
 
 	logout("startup sequence complete\n");
 
@@ -604,7 +697,7 @@ main(int argc, char * const argv[])
 	/* since workers will be freed, stop querying the structures */
 	collector_stop();
 	server_shutdown(internal_submission);
-	free(internal_submission);
+	server_free(internal_submission);
 	logout("stopped collector\n");
 	if (numaggregators > 0) {
 		aggregator_stop();
@@ -619,6 +712,7 @@ main(int argc, char * const argv[])
 		dispatch_stop(workers[id + 0]);
 	for (id = 0; id < 1 + workercnt; id++) {
 		dispatch_shutdown(workers[id + 0]);
+		dispatch_free(workers[id + 0]);
 		fprintf(relay_stdout, " %d", id + 1);
 		fflush(relay_stdout);
 	}
@@ -626,14 +720,12 @@ main(int argc, char * const argv[])
 	fflush(relay_stdout);
 	free(workers);
 
-	router_shutdown();
-	servers = router_getservers(clusters);
-	for (i = 0; servers[i] != NULL; i++)
-		server_stop(servers[i]);
-	free(servers);
-	router_free(clusters, routes);
+	router_free(rtr);
 	logout("stopped servers\n");
 
+	if (pidfile != NULL)
+		unlink(pidfile);
+
 	logout("stopped carbon-c-relay v%s (%s)\n", VERSION, GIT_VERSION);
 
 	return 0;
