@@ -29,6 +29,7 @@
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <sys/resource.h>
+#include <assert.h>
 
 #include "relay.h"
 #include "queue.h"
@@ -71,7 +72,7 @@ struct _server {
  * batched, but sent one by one to reduce loss on sending failure.
  * A connection with the server is maintained for as long as there is
  * data to be written.  As soon as there is none, the connection is
- * dropped.
+ * dropped if a timeout of DISCONNECT_WAIT_TIME exceeds.
  */
 static void *
 server_queuereader(void *d)
@@ -82,8 +83,7 @@ server_queuereader(void *d)
 	const char **metric = self->batch;
 	struct timeval start, stop;
 	struct timeval timeout;
-	int timeoutms;
-	queue *queue;
+	queue *squeue;
 	char idle = 0;
 	size_t *secpos = NULL;
 
@@ -164,14 +164,14 @@ server_queuereader(void *d)
 			 * - do nothing (we will overflow, since we can't send
 			 *   anywhere) */
 			*metric = NULL;
-			queue = NULL;
+			squeue = NULL;
 			for (i = 0; i < self->secondariescnt; i++) {
 				/* both conditions below make sure we skip ourself */
 				if (self->secondaries[secpos[i]]->failure)
 					continue;
-				queue = self->secondaries[secpos[i]]->queue;
-				if (!self->failover && LEN_CRITICAL(queue)) {
-					queue = NULL;
+				squeue = self->secondaries[secpos[i]]->queue;
+				if (!self->failover && LEN_CRITICAL(squeue)) {
+					squeue = NULL;
 					continue;
 				}
 				if (*metric == NULL) {
@@ -183,7 +183,7 @@ server_queuereader(void *d)
 				}
 
 				for (; *metric != NULL; metric++)
-					if (!queue_putback(queue, *metric))
+					if (!queue_putback(squeue, *metric))
 						break;
 				/* try to put back stuff that didn't fit */
 				for (; *metric != NULL; metric++)
@@ -191,14 +191,14 @@ server_queuereader(void *d)
 						break;
 			}
 			for (; *metric != NULL; metric++) {
-				if (mode == DEBUG)
+				if (mode & MODE_DEBUG)
 					logerr("dropping metric: %s", *metric);
 				free((char *)*metric);
 				self->dropped++;
 			}
 			gettimeofday(&stop, NULL);
 			self->ticks += timediff(start, stop);
-			if (queue == NULL) {
+			if (squeue == NULL) {
 				/* we couldn't do anything, take it easy for a bit */
 				if (self->failure)
 					self->failure = 1;
@@ -338,10 +338,10 @@ server_queuereader(void *d)
 				(void) fcntl(self->fd, F_SETFL, args);
 			}
 
-			/* ensure we will break out of connections being stuck */
-			timeoutms = self->iotimeout + (rand() % 100);
-			timeout.tv_sec = timeoutms / 1000;
-			timeout.tv_usec = (timeoutms % 1000) * 1000;
+			/* ensure we will break out of connections being stuck more
+			 * quickly than the kernel would give up */
+			timeout.tv_sec = 10;
+			timeout.tv_usec = (rand() % 300) * 1000;
 			setsockopt(self->fd, SOL_SOCKET, SO_SNDTIMEO,
 					&timeout, sizeof(timeout));
 #ifdef SO_NOSIGPIPE
@@ -390,7 +390,7 @@ server_queuereader(void *d)
 				/* put back stuff we couldn't process */
 				for (; *metric != NULL; metric++) {
 					if (!queue_putback(self->queue, *metric)) {
-						if (mode == DEBUG)
+						if (mode & MODE_DEBUG)
 							logerr("server %s:%u: dropping metric: %s",
 									self->ip, self->port, *metric);
 						free((char *)*metric);
@@ -487,8 +487,7 @@ server_new(
 /**
  * Adds a list of secondary servers to this server.  A secondary server
  * is a server which' queue will be checked when this server has nothing
- * to do.  This is different from a backup server in that all servers
- * involved have their own queue which they are supposed to deal with.
+ * to do.
  */
 void
 server_add_secondaries(server *self, server **secondaries, size_t count)
@@ -551,16 +550,6 @@ server_send(server *s, const char *d, char force)
 	return 1;
 }
 
-/**
- * Signals this server to stop whatever it's doing.
- */
-void
-server_stop(server *s)
-{
-	if (s->secondariescnt == 0)
-		s->keep_running = 0;
-}
-
 /**
  * Waits for this server to finish sending pending items from its queue.
  */
@@ -568,17 +557,12 @@ void
 server_shutdown(server *s)
 {
 	int i;
-	pthread_t tid;
 	size_t failures;
 	size_t inqueue;
 	int err;
-	const char *p;
 
-	if (s->tid == 0)
-		return;
-
-	tid = s->tid;
-	s->tid = 0;
+	/* this function should only be called once for each server */
+	assert(s->tid != 0);
 
 	if (s->secondariescnt > 0) {
 		/* if we have a working connection, or we still have stuff in
@@ -596,7 +580,7 @@ server_shutdown(server *s)
 			/* loop until we all failed, or nothing is in the queues */
 		} while (failures != s->secondariescnt &&
 				inqueue != 0 &&
-				logerr("any_of cluster pending %zu metrics "
+				logout("any_of cluster pending %zu metrics "
 					"(with %zu failed nodes)\n", inqueue, failures) >= -1 &&
 				usleep((200 + (rand() % 100)) * 1000) <= 0);
 		/* shut down entire cluster */
@@ -609,9 +593,10 @@ server_shutdown(server *s)
 	}
 
 	s->keep_running = 0;
-	if ((err = pthread_join(tid, NULL)) != 0)
+	if ((err = pthread_join(s->tid, NULL)) != 0)
 		logerr("%s:%u: failed to join server thread: %s\n",
 				s->ip, s->port, strerror(err));
+	s->tid = 0;
 
 	if (s->ctype == CON_TCP) {
 		size_t qlen = queue_len(s->queue);
@@ -619,9 +604,13 @@ server_shutdown(server *s)
 			logerr("dropping %zu metrics for %s:%u\n",
 					qlen, s->ip, s->port);
 	}
-	/* drain queue not to leak the memory consumed by pending metrics */
-	while ((p = queue_dequeue(s->queue)) != NULL)
-		free((char *)p);
+}
+
+/**
+ * Frees this server and associated resources.
+ */
+void
+server_free(server *s) {
 	queue_destroy(s->queue);
 	free(s->batch);
 	if (s->instance)
@@ -630,6 +619,7 @@ server_shutdown(server *s)
 		freeaddrinfo(s->saddr);
 	free((char *)s->ip);
 	s->ip = NULL;
+	free(s);
 }
 
 /**
